
#include <math.h>
#include <malloc.h>
#include <string.h>

#include "hashtab.h"

uint Hash ( byte *key, uint keySize, uint hashMax )
{
  int i;
  double sum=0;
  byte *p;

  for (i=0, p=key; i<keySize; p++, i++) {
    sum += *p;
  }
  if (sum<0) sum = -sum;
  return ((uint)sum) % hashMax;
}

HashTab :: HashTab()
{
  buckets = NULL;
  ovfArea = nextOvf = NULL;
  bufferArea = NULL;
  bufferSize = ovfAreaSize = nBuckets = objSize = keyOffs = keySize = 0;
  bucketSize = bucketCapacity = 0;
}

HashTab :: HashTab ( byte * buf, uint bufSize, byte * ovf, uint ovfSize,
          uint bucks, uint oSize, uint kOffs, uint kSize )
{
  HashTab();
  Setup ( buf, bufSize, ovf, ovfSize, bucks, oSize, kOffs, kSize );
}

HashTab :: ~HashTab()
{
  Destroy();
}

int HashTab :: Setup ( byte * buf, uint bufSize, byte * ovf, uint ovfSize,
                        uint bucks, uint oSize, uint kOffs, uint kSize )
{
  int    i, bs;
  Bucket * b;
  byte   * p;

  if (buckets)      // already allocated. First use Destroy().
    return ERROR;

  buckets = (Bucket *) malloc ( sizeof(Bucket) * bucks );
  if (buckets == NULL)
    return ERROR;

  bufferArea  = buf;
  bufferSize  = bufSize;
  ovfArea     = nextOvf = ovf;
  ovfAreaSize = ovfSize;
  nBuckets    = bucks;
  objSize     = oSize;
  keyOffs     = kOffs;
  keySize     = kSize;

  bucketSize  = bufSize / bucks;
  bucketCapacity = bucketSize / oSize;

  if (buf+bufSize == ovf) {             // buf & ovf are adjecent; resize them
    int oldBufSize = bufSize;
    bucketSize = bucketCapacity * oSize;
    bufSize = bucks * bucketSize;
    ovf = buf + bufSize;
    ovfSize += bufSize - oldBufSize;
  }

  for ( i=0, p=buf, b=buckets; i<bucks; i++, p+=bucketSize, b++ ) {
    b->curr  = b->start = p;
    b->ovf   = NULL;
    b->avail = bucketCapacity;
  }

  return OK;
}

void HashTab :: Destroy()
{
  if (buckets) free(buckets);

  buckets = NULL;
  ovfArea = NULL;
  bufferArea = NULL;
  bufferSize = ovfAreaSize = nBuckets = objSize = keyOffs = keySize = 0;
  bucketSize = bucketCapacity = 0;
}

int HashTab :: Put(byte *data)
{
  int addr;
  Bucket * b;

  addr = Hash ( data+keyOffs, keySize, nBuckets );
  b = &buckets[addr];

  while (true)

    if (b->avail) {   // Trivial case

      memcpy(b->curr, data, objSize);
      b->curr += objSize;
      b->avail--;
      return OK;

    } else                        // Overflow occured

       if (b->ovf)                // An ovf area was allocated before

         b = b->ovf;              // recurse on the overflow bucket

       else {                     // This is the first overflow at this bucket

         if (ovfAreaSize < b->avail * objSize) return ERROR; // No ovf space

         b->ovf = (Bucket *) malloc (sizeof(Bucket));
         if (b->ovf == NULL) return ERROR;

         b = b->ovf;
         b->curr  = b->start = ovfArea;
         b->ovf   = NULL;
         b->avail = bucketCapacity / 2;  // Overflow buckets are half the size of others
         if (!b->avail)
            b->avail = 1;

         ovfArea += b->avail * objSize;
         ovfAreaSize -= b->avail * objSize;

       }
}

byte * HashTab :: Search (byte *data, Bucket **bp, int *n)
// data : IN - pointer to the object info being searched
// bp   : OUT - address of a pointer to the bucket in which the object is found
// n    : OUT - the position of the object in the bucket
{
  int i, addr, bc;
  Bucket * b;
  byte *p;

  addr = Hash ( data+keyOffs, keySize, nBuckets );
  b = &buckets[addr];
  bc = bucketCapacity;

  while(1) {
    for (i=0, p=b->start; i<bc-b->avail; p+=objSize, i++)
      if (memcmp(data+keyOffs, p+keyOffs, keySize) == 0) {
         *n = i;
         *bp = b;
         return p;
      }
    if (b->ovf) { b=b->ovf; bc=(b->curr - b->start) / objSize + b->avail; }
    else return NULL;
  }
}

byte * HashTab :: SearchNext (byte *data, Bucket **bp, int *n)
{
  int i, addr, bc;
  Bucket * b;
  byte *p;

  b = *bp;
  bc = (b->curr - b->start) / objSize + b->avail;  // capacity of this bucket
  p = b->start + (*n + 1) * objSize;
  i = *n +1;
  while(1) {
    for (; i<bc-b->avail; p+=objSize, i++)
      if (memcmp(data+keyOffs, p+keyOffs, keySize) == 0) {
         *n = i;
         *bp = b;
         return p;
      }
    if (b->ovf) {
      b = b->ovf;
      bc = (b->curr - b->start) / objSize + b->avail;
      p = b->start;
      i = 0;
    } else return NULL;
  }
}
